import json
import os
import typing

from autochat.base import AutochatBase
from autochat.model import Message, MessagePart
from autochat.providers.base_provider import BaseProvider
from autochat.providers.utils import add_empty_function_result

AUTOCHAT_HOST = os.getenv("AUTOCHAT_HOST")


class FunctionCallParsingError(Exception):
    def __init__(self, id, function_call):
        self.id = id
        self.function_call = function_call

    def __str__(self):
        return f"Invalid function_call: {self.obj.function_call}"


def from_openai_object(
    role: str,
    content: str,
    tool_calls: typing.Optional[list] = None,
    id: typing.Optional[str] = None,
):
    # We need to unquote the arguments

    function_call_dict = None
    function_call_id = None
    if tool_calls:
        first_tool_call_function = next(
            (t for t in tool_calls if t.type == "function"), None
        )
        if first_tool_call_function is None:
            raise ValueError(
                "We don't support tool calls with type other than function"
            )
        function_call = first_tool_call_function.function
        function_call_id = first_tool_call_function.id
        try:
            function_call_dict = {
                "name": function_call.name,
                "arguments": json.loads(function_call.arguments),
            }
        except json.decoder.JSONDecodeError:
            raise FunctionCallParsingError(id, function_call)

    return Message(
        role=role,
        content=content,
        function_call=function_call_dict,
        id=id,
        function_call_id=function_call_id,
    )


def parts_to_openai_dict(part: MessagePart) -> dict:
    if part.type == "text":
        return {
            "type": "text",
            "text": part.content,
        }
    elif part.type == "image":
        return {
            "type": "image_url",
            "image_url": {
                "url": f"data:{part.image.format};base64,{part.image.to_base64()}"
            },
        }
    elif part.type == "function_call":
        return {
            "name": part.function_call["name"],
            "arguments": json.dumps(part.function_call["arguments"]),
        }
    elif part.type == "function_result":
        return {
            "type": "text",
            "text": part.content,
        }
    elif part.type == "function_result_image":
        return {
            "type": "image_url",
            "image_url": {
                "url": f"data:{part.image.format};base64,{part.image.to_base64()}"
            },
        }

    raise ValueError(f"Unknown part type: {part.type}")


def message_to_openai_dict(message: Message) -> dict:
    if message.role == "function":
        res = {
            "role": "tool",
            "tool_call_id": message.function_call_id,
            "content": [parts_to_openai_dict(part) for part in message.parts],
            "name": message.name,
        }
    else:
        res = {"role": message.role, "content": []}
        for part in message.parts:
            if part.type == "function_call" and message.role == "assistant":
                res["tool_calls"] = [
                    {
                        "id": part.function_call_id,
                        "type": "function",
                        "function": parts_to_openai_dict(part),
                    }
                ]
            elif part.type == "function_call" and message.role == "user":
                # Workaround: If the user is triggering a function, we add it's name and arguments to the content
                res["content"] = (
                    part.function_call["name"]
                    + ":"
                    + json.dumps(part.function_call["arguments"])
                )
            else:
                res["content"].append(parts_to_openai_dict(part))

        if "content" in res and len(res["content"]) == 0:
            res["content"] = None

    return res


def return_image_as_user_message(messages: list[Message]) -> list[Message]:
    """
    Adaptation because OpenAI doesn't support image in function call.
    We return the image as a user message.
    """
    for i in range(len(messages)):
        if messages[i].role == "function" and messages[i].image is not None:
            messages[i].role = "user"
    return messages


class OpenAIProvider(BaseProvider):
    def __init__(self, chat: AutochatBase, model: str, base_url: str = None):
        from openai import OpenAI

        self.chat = chat
        self.model = model
        # Possibly set up openai.api_key, base_url, etc.
        self.client = OpenAI(
            base_url=(AUTOCHAT_HOST if AUTOCHAT_HOST else "https://api.openai.com/v1"),
        )

    async def fetch_async(self, **kwargs) -> Message:
        messages = self.prepare_messages(
            transform_function=message_to_openai_dict,
            transform_list_function=lambda x: add_empty_function_result(
                return_image_as_user_message(x)
            ),
        )

        system_messages = []
        if self.chat.instruction:
            system_messages.append(
                Message(
                    role="system",
                    content=self.chat.instruction,
                )
            )
        if self.chat.last_tools_states:
            system_messages.append(
                Message(
                    role="system",
                    content=self.chat.last_tools_states,
                )
            )
        messages = [message_to_openai_dict(sm) for sm in system_messages] + messages

        if self.chat.use_tools_only and "tool_choice" not in kwargs:
            kwargs["tool_choice"] = "required"

        if self.chat.functions_schema:
            tools = [
                {
                    "type": "function",
                    "function": tool,
                }
                for tool in self.chat.functions_schema
            ]
            res = self.client.chat.completions.create(
                model=self.model,
                messages=messages,
                tools=tools,
                **kwargs,
            )
        else:
            res = self.client.chat.completions.create(
                model=self.model, messages=messages, **kwargs
            )

        message = res.choices[0].message
        return from_openai_object(
            role=message.role,
            content=message.content,
            tool_calls=message.tool_calls,
            id=res.id,  # We use the response id as the message id
        )



import os

from autochat.model import Message
from autochat.providers.base_provider import APIProvider, BaseProvider


# TODO: should probably exploit default model from provider
def get_provider_and_model(  # TODO: get_provider_and_model ?
    # chat: Autochat, # TODO: make AutochatBase ?
    chat,
    provider_name: str = None,
    model: str = None,
) -> list[str, BaseProvider]:  # TODO: rename
    """
    Returns the correct LLM provider based on a string or env vars.
    """
    from autochat.providers.anthropic import AnthropicProvider
    from autochat.providers.openai import OpenAIProvider
    from autochat.providers.openai_function_legacy import OpenAIProviderFunctionLegacy
    from autochat.providers.openai_function_shim import OpenAIProviderFunctionShim

    if not provider_name:
        provider_name = os.getenv("AUTOCHAT_HOST", "openai")

    if isinstance(provider_name, APIProvider):
        provider_key = provider_name
    elif isinstance(provider_name, str):
        try:
            provider_key = APIProvider(provider_name)
        except ValueError:
            raise ValueError(f"Provider {provider_name} is not a valid provider")
    else:
        raise ValueError(f"Invalid provider: {provider_name}")

    if provider_key == APIProvider.OPENAI:
        if not model:
            model = os.getenv("AUTOCHAT_MODEL", "gpt-4o")
        return OpenAIProvider(chat, model=model), model
    elif provider_key == APIProvider.ANTHROPIC:
        if not model:
            model = os.getenv("AUTOCHAT_MODEL", "claude-3-7-sonnet-latest")
        return AnthropicProvider(chat, model=model), model
    elif provider_key == APIProvider.OPENAI_FUNCTION_SHIM:
        if not model:
            model = os.getenv("AUTOCHAT_MODEL", "o1-preview")
        return OpenAIProviderFunctionShim(chat, model=model), model
    elif provider_key == APIProvider.OPENAI_FUNCTION_LEGACY:
        if not model:
            model = os.getenv("AUTOCHAT_MODEL", "gpt-4o")
        return OpenAIProviderFunctionLegacy(chat, model=model), model
    else:
        raise ValueError(f"Provider {provider_key} is not supported")


def add_empty_function_result(messages: list[Message]) -> list[Message]:
    """
    OpenAI/Anthropic requires a call to have a result message. Not what we want.
    Adjustment for merging or inserting the "function_result":

    - First case: a message with `role="function"` followed by a message with `role="user"`.
      We transform this 'function' message into a part of type 'tool_result' and insert it at the beginning of the following user message.

    - Second case (unchanged): a message with `role="assistant"` containing a `function_call`, followed by a non-`function` message.
      We insert an empty message with `role="function"`.
    """
    for i in range(len(messages) - 1, 0, -1):
        if (
            messages[i - 1].role == "assistant"
            and messages[i - 1].function_call
            and not messages[i].role == "function"
        ):
            # Insert an empty function result
            messages.insert(
                i,
                Message(
                    role="function",
                    name=messages[i - 1].function_call["name"],
                    content="",
                    function_call_id=messages[i - 1].function_call_id,
                ),
            )
    return messages



"""
Temporary provider so we can use function calls with OpenAI models
that don't support them yet in the API (ie. o1).
"""

import json
import typing

from autochat.model import Message
from autochat.providers.openai import OpenAIProvider, parts_to_openai_dict


def parse_function_call_from_text(text: str):
    import re

    """
    Look for something like:

    CALLING FUNCTION: my_function
    {
        "foo": "bar"
    }

    Return (function_name, arguments_dict) or (None, None) if no function call found.
    This is a simplistic regex approach. Adjust as needed.
    """
    pattern = r"(?s)CALLING\s+FUNCTION:\s*([A-Za-z0-9_]+)\s*\n?\s*(\{.*?\})"
    match = re.search(pattern, text)
    if match:
        function_name = match.group(1).strip()
        try:
            # Attempt to parse JSON in the curly braces
            arguments = json.loads(match.group(2))
        except json.JSONDecodeError:
            arguments = {}
        return function_name, arguments
    return None, None


def from_openai_object(
    role: str,
    content: str,
    id: typing.Optional[str] = None,
):
    function_name, function_args = parse_function_call_from_text(content)
    function_call = None
    if function_name:
        function_call = {
            "name": function_name,
            "arguments": function_args,
        }
        # Shortcut
        content = None
    return Message(role=role, content=content, function_call=function_call, id=id)


def message_to_openai_dict(message: Message) -> dict:
    """o1 only supports text-based instructions for function usage"""
    res = {
        "role": "user" if message.role == "function" else message.role,
        "content": [],
    }

    for part in message.parts:
        if part.type == "function_call":
            text = f"CALLING FUNCTION: {part.function_call['name']}\n"
            text += f"{json.dumps(part.function_call['arguments'])}\n"
            res["content"].append({"type": "text", "text": text})
        else:
            res["content"].append(parts_to_openai_dict(part))

    return res


class OpenAIProviderFunctionShim(OpenAIProvider):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    async def fetch_async(self):
        """
        Calls the OpenAI ChatCompletion endpoint using ONLY text-based instructions for function usage.
        """

        messages = self.prepare_messages(transform_function=message_to_openai_dict)
        # Add instruction as the first message
        if self.chat.instruction:
            instruction_message = Message(
                role="user",
                content=self.chat.instruction,
            )
            messages = [message_to_openai_dict(instruction_message)] + messages

        # Inject function schema into the system message, if we haven't already
        # or just append it to an existing system message. Example:
        if self.chat.functions_schema:
            function_schemas_text = json.dumps(self.chat.functions_schema, indent=2)
            augmented_instruction = (
                (self.chat.instruction or "")
                + "\n\nHere are the available functions you can call, in JSON format:\n"
                + function_schemas_text
                + "\n\n"
                + "When you call a function, respond with:\n"
                + "```\n"
                + "CALLING FUNCTION: <function_name>\n"
                + "{\n"
                + '  "arg1": "...",\n'
                + '  "arg2": "..."\n'
                + "}\n"
                + "```\n"
                + "If you do not need to call a function, just respond normally."
            )
        else:
            augmented_instruction = self.chat.instruction

        if augmented_instruction:
            system_msg = Message(
                role="user",
                content=augmented_instruction,
            )
            messages = [message_to_openai_dict(system_msg)] + messages

        # Make the request WITHOUT 'functions='
        res = self.client.chat.completions.create(
            model=self.model,
            messages=messages,
        )
        # OpenAI returns a structure. We'll parse out the text content
        message = res.choices[0].message
        # Note that now there is no function_call field because we didn't pass `functions=`.
        return from_openai_object(
            role=message.role,
            content=message.content,
            # no function_call here
            id=res.id,
        )



from autochat.chat import Autochat


class DummyTool:
    def method1(self):
        pass

    def method2(self, param: str):
        pass

    def _private_method(self):
        pass


def test_add_tool():
    agent = Autochat()
    tool_id = agent.add_tool(DummyTool)

    # Check if the tool was added correctly
    assert tool_id in agent.tools
    assert isinstance(agent.tools[tool_id], DummyTool)

    # Check if all methods (excluding private) were added to functions_schema
    function_names = [f["name"] for f in agent.functions_schema]
    assert f"DummyTool-{tool_id}__method1" in function_names
    assert f"DummyTool-{tool_id}__method2" in function_names
    assert f"DummyTool-{tool_id}___private_method" not in function_names

    # Check if all functions (excluding private) were added correctly
    assert f"DummyTool-{tool_id}__method1" in agent.functions
    assert f"DummyTool-{tool_id}__method2" in agent.functions
    assert f"DummyTool-{tool_id}___private_method" not in agent.functions


def test_add_tool_with_custom_id():
    agent = Autochat()
    custom_id = "custom_tool_id"
    tool_id = agent.add_tool(DummyTool(), tool_id=custom_id)

    assert tool_id == custom_id
    assert custom_id in agent.tools
    assert isinstance(agent.tools[custom_id], DummyTool)


def test_add_tool_instance():
    agent = Autochat()
    tool_instance = DummyTool()
    tool_id = agent.add_tool(tool_instance)

    assert tool_id in agent.tools
    assert agent.tools[tool_id] is tool_instance

    # Check if all methods (excluding private) were added to functions_schema
    function_names = [f["name"] for f in agent.functions_schema]
    assert f"DummyTool-{tool_id}__method1" in function_names
    assert f"DummyTool-{tool_id}__method2" in function_names
    assert f"DummyTool-{tool_id}___private_method" not in function_names



"""AutoChat package."""

__version__ = "0.4.1"

import inspect
import io
import json
import logging
import os
import traceback
import typing
import warnings

from PIL import Image as PILImage

from autochat.base import AutochatBase
from autochat.model import Message
from autochat.providers.base_provider import APIProvider
from autochat.providers.utils import get_provider_and_model
from autochat.utils import (
    csv_dumps,
    get_event_loop_or_create,
    inspect_schema,
    parse_chat_template,
)

AUTOCHAT_HOST = os.getenv("AUTOCHAT_HOST")
AUTOCHAT_MODEL = os.getenv("AUTOCHAT_MODEL")
OUTPUT_SIZE_LIMIT = int(os.getenv("AUTOCHAT_OUTPUT_SIZE_LIMIT", 4000))


class StopLoopException(Exception):
    pass


def simple_response_default_callback(response: Message) -> Message:
    """
    This function is called when the response is a simple response (no function call).
    By default, the conversation will stop after a simple response.
    You can override this function to change this behavior.
    """
    raise StopLoopException("Stopping the conversation after a simple response")


class Autochat(AutochatBase):
    def __init__(
        self,
        instruction: str = None,
        examples: typing.Union[list[Message], None] = None,
        messages: typing.Union[list[Message], None] = None,
        context: str = None,
        max_interactions: int = 100,
        model=AUTOCHAT_MODEL,
        provider: str = APIProvider.OPENAI,
        use_tools_only: bool = False,
        mcp_servers: typing.Union[list[object], None] = [],
    ) -> None:
        """
        Initialize the Autochat instance.
        Args:
            use_tools_only: bool = False,
                If True, the chat will only use tools and not the LLM.
                This is a beta feature and may change in the future.
        """
        self.provider, self.model = get_provider_and_model(
            self, provider, model
        )  # TODO: rename register ?
        self.simple_response_callback = simple_response_default_callback
        if use_tools_only:
            warnings.warn(
                "use_tools_only is a beta feature and may change in the future"
            )
        self.use_tools_only = use_tools_only
        self.client = None  # TODO:
        self.instruction = instruction
        if examples is None:
            self.examples = []
        else:
            self.examples = examples

        if messages is None:
            self.messages = []
        else:
            self.messages = messages

        self.context = context
        self.max_interactions = max_interactions
        self.functions_schema = []
        self.functions = {}
        self.tools = {}
        for m in mcp_servers:
            self.add_mcp_server(m)

    @classmethod
    def from_template(cls, chat_template: str, **kwargs):
        instruction, examples = parse_chat_template(chat_template)
        return cls(
            instruction=instruction,
            examples=examples,
            **kwargs,
        )

    def reset(self):
        """Reset the chat state.

        This will clear the chat history and reset the functions and tools.
        """
        self.messages = []
        self.functions_schema = []
        self.functions = {}
        self.tools = {}

    @property
    def last_message(self):
        if not self.messages:
            return None
        return self.messages[-1].content

    @property
    def last_tools_states(self) -> typing.Optional[str]:
        """We add the repr() of each tool to the system context"""
        # If there are no tools, return None
        if not self.tools:
            return None
        tool_reprs = []
        for tool_id, tool in self.tools.items():
            tool_name = f"{tool.__class__.__name__}-{tool_id}"
            if hasattr(tool, "__repr__"):
                tool_reprs.append(f"### {tool_name}\n{repr(tool)}")

        tool_context = "\n".join(tool_reprs)

        """
        ## Last Tools States
        ### Tool 1
        {state}
        ### Tool 2
        {state}
        --- End of Last Tools States ---
        """
        return f"## Last Tools States\n{tool_context}\n--- End of Last Tools States ---"

    def load_messages(self, messages: list[Message]):
        # Check order of messages (based on createdAt)
        # Oldest first (createdAt ASC)
        # messages = sorted(messages, key=lambda x: x.createdAt)
        self.messages = messages  # [message for message in messages]

    def add_function(
        self,
        function: typing.Callable,
        function_schema: typing.Optional[dict] = None,
    ):
        if function_schema is None:
            # We try to infer the function schema from the function
            function_schema = inspect_schema(function)

        self.functions_schema.append(function_schema)
        self.functions[function_schema["name"]] = function

    def add_tool(
        self, tool: typing.Union[type, object], tool_id: typing.Optional[str] = None
    ) -> str:
        """Add a tool class or instance to the chat instance and return the tool id"""
        if isinstance(tool, type):
            # If a class is provided, instantiate it
            tool = tool()

        if tool_id is None:
            tool_id = str(id(tool))

        self.tools[tool_id] = tool

        class_name = tool.__class__.__name__
        tool_name = f"{class_name}-{tool_id}"
        for method_name, method in inspect.getmembers(tool, inspect.ismethod):
            if method_name.startswith("_"):  # Skip private methods
                continue
            function_schema = inspect_schema(method)
            function_schema["name"] = f"{tool_name}__{method_name}"
            self.add_function(method, function_schema)
        return tool_id

    def remove_tool(self, tool_id: str):
        del self.tools[tool_id]
        self.functions_schema = [
            schema for schema in self.functions_schema if schema["name"] != tool_id
        ]
        self.functions = {
            name: func for name, func in self.functions.items() if name != tool_id
        }

    async def add_mcp_server(self, mcp_server):
        """
        Add a MCP server to the agent instance.
        The MCP server will be used to call tools and resources.
        """
        from autochat.mcp import fetch_mcp_server_resources, fetch_mcp_server_tools

        logging.warning("Experimental feature: MCP servers")
        tools_functions, schemas = await fetch_mcp_server_tools(mcp_server)
        self.functions.update(tools_functions)
        self.functions_schema.extend(schemas)
        resources_functions, schemas = await fetch_mcp_server_resources(mcp_server)
        self.functions.update(resources_functions)
        self.functions_schema.extend(schemas)

    async def ask_async(
        self,
        message: typing.Union[Message, str, None] = None,
        **kwargs,
    ) -> Message:
        """Async version of ask method that uses the async fetch_async provider method"""
        if message:
            if isinstance(message, str):
                # If message is instance of string, then convert to Message
                message = Message(
                    role="user",
                    content=message,
                )
            self.messages.append(message)  # Add the question to the history

        # Call the async strategy
        response = await self.provider.fetch_async(**kwargs)
        self.messages.append(response)
        return response

    def ask(
        self,
        message: typing.Union[Message, str, None] = None,
        **kwargs,
    ) -> Message:
        # For backward compatibility, use run_until_complete to execute the async method
        loop = get_event_loop_or_create()
        return loop.run_until_complete(self.ask_async(message, **kwargs))

    def _format_callback_message(
        self, function_name, function_call_part_id, content, image
    ):
        if isinstance(content, Message):
            message = content
            # TODO: Add name and function_call_id to the message
            # message.name = function_name
            # message.function_call_id = function_call_part_id
            return message

        # We format the function_call response to be used as a next message
        if content is None:
            # When content is None, use an empty string instead to prevent the "Message should have at least one part" error
            formatted_content = ""
        elif isinstance(content, list):
            formatted_content = []
            if not content:
                formatted_content = "[]"
            elif isinstance(content[0], dict):
                # If data is list of dicts, dumps to CSV
                formatted_content = csv_dumps(content, OUTPUT_SIZE_LIMIT)
            else:
                for item in content:
                    if isinstance(item, str):
                        formatted_content.append(item)
                    elif isinstance(item, object):
                        tool_id = self.add_tool(item)
                        formatted_content.append(f"Added tool: {tool_id}")
                    elif isinstance(item, (int, float, bool)):
                        formatted_content.append(str(item))
                    else:
                        raise ValueError(f"Invalid item type: {type(item)}")
            formatted_content = "\n".join(formatted_content)
            # Limit the size of the content
            if len(formatted_content) > OUTPUT_SIZE_LIMIT:
                formatted_content = (
                    formatted_content[:OUTPUT_SIZE_LIMIT]
                    + f"\n... ({len(formatted_content)} characters)"
                )
        elif isinstance(content, dict):
            content_dump = json.dumps(content)
            if len(content_dump) > OUTPUT_SIZE_LIMIT:
                formatted_content = (
                    content_dump[:OUTPUT_SIZE_LIMIT]
                    + f"\n... ({len(content_dump)} characters)"
                )
            else:
                formatted_content = content_dump
        elif isinstance(content, str):
            if len(content) > OUTPUT_SIZE_LIMIT:
                formatted_content = (
                    content[:OUTPUT_SIZE_LIMIT] + f"\n... ({len(content)} characters)"
                )
            else:
                formatted_content = content
        elif isinstance(content, bytes):
            # Detect if it's an image
            try:
                image = PILImage.open(io.BytesIO(content))
                formatted_content = None
            except IOError:
                # Not an image
                raise ValueError("Returned bytes is not a valid image.")
        elif isinstance(content, PILImage.Image):
            image = content
            formatted_content = None
        elif isinstance(content, (int, float, bool)):
            formatted_content = str(content)
        elif isinstance(content, object):
            # If the function return an object, we add it as a tool
            # NOTE: Maybe we shouldn't, and rely on a clearer signal / object type ?
            tool_id = self.add_tool(content)
            formatted_content = f"Added tool: {tool_id}"
        else:
            raise ValueError(f"Invalid content type: {type(content)}")

        # Build a function response message
        return Message(
            name=function_name,
            role="function",
            content=formatted_content,
            image=image,
            function_call_id=function_call_part_id,
        )

    def _format_exception(self, e):
        # We clean the traceback to remove frames from __init__.py
        tb = traceback.extract_tb(e.__traceback__)
        filtered_tb = [frame for frame in tb if "chat.py" not in frame.filename]
        if filtered_tb:
            content = "Traceback (most recent call last):\n"
            content += "".join(traceback.format_list(filtered_tb))
            content += f"\n{e.__class__.__name__}: {str(e)}"
        else:
            # If no relevant frames, use the full traceback
            content = traceback.format_exc()
        return content

    async def _call_with_signature(self, func, from_response, **kwargs):
        sig = inspect.signature(func)
        if "from_response" in sig.parameters:
            result = func(**kwargs, from_response=from_response)
        else:
            result = func(**kwargs)

        # If the function is async, await the result
        if inspect.iscoroutine(result):
            return await result
        else:
            return result

    async def _call_function_and_build_message(
        self, function_name, function_arguments, response
    ):
        """
        Encapsulate the 'call the function' logic & handle exceptions
        plus building a function or tool response message.
        """
        content = None
        image = None

        try:
            if function_name.startswith("tool-"):
                tool_id, method_name = function_name.split("__")
                tool = self.tools[tool_id]
                method = getattr(tool, method_name)
                content = await self._call_with_signature(
                    method, response, **function_arguments
                )
            else:
                content = await self._call_with_signature(
                    self.functions[function_name], response, **function_arguments
                )
        except StopLoopException:
            raise
        except Exception as e:
            content = self._format_exception(e)

        # Build next message
        return self._format_callback_message(
            function_name=function_name,
            function_call_part_id=response.function_call_id,
            content=content,
            image=image,
        )

    async def run_conversation_async(
        self,
        question: typing.Union[str, Message, None] = None,
    ) -> typing.AsyncGenerator[Message, None]:
        """Async version of run_conversation"""
        # If there's an initial question, emit it:
        if isinstance(question, str):
            message = Message(
                role="user",
                content=question,
            )
            yield message
        else:
            message = question

        for _ in range(self.max_interactions):
            # Ask the LLM with the current message (if any)
            response = await self.ask_async(message)

            # Locate a function_call part in the assistant's response
            function_call_part = next(
                (p for p in response.parts if p.type == "function_call"), None
            )
            if not function_call_part:
                yield response
                try:
                    message = self.simple_response_callback(response)
                except StopLoopException:
                    return
            else:
                name = function_call_part.function_call["name"]
                args = function_call_part.function_call["arguments"]

                try:
                    message = await self._call_function_and_build_message(
                        name, args, response
                    )
                except StopLoopException:
                    return
                finally:
                    yield response
                yield message

    def run_conversation(self, question: typing.Union[str, Message, None] = None):
        # For backward compatibility, use run_until_complete to execute the async method
        loop = get_event_loop_or_create()
        async_gen = self.run_conversation_async(question)

        while True:
            try:
                # Get the next item from the async generator
                item = loop.run_until_complete(async_gen.__anext__())
                # Yield the item immediately
                yield item
            except StopAsyncIteration:
                # Stop when the async generator is exhausted
                break



"""We want to be able to return an image from a function call."""

import os

import pytest
from PIL import Image

from autochat import APIProvider, Autochat

img_path = os.path.join(os.path.dirname(__file__), "images", "mileage.jpg")


def read_image():
    return Image.open(img_path)


@pytest.mark.parametrize(
    "provider",
    [
        APIProvider.OPENAI_FUNCTION_LEGACY,
        APIProvider.ANTHROPIC,
    ],
)
@pytest.mark.vcr
def test_anthropic_callback_image(provider):
    agent = Autochat(
        provider=provider,
    )
    agent.add_function(read_image)

    user_query = "Hello, read the image and explain it"
    messages = []
    for message in agent.run_conversation(user_query):
        messages.append(message)
